/*
 * if_u32.c - Processing of the "if" command, using the "u32" classifier
 *
 * Written 2001,2002 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 * Copyright 2002 Network Robots
 */


#include <stdint.h>
#include <netinet/in.h>

#include "config.h"
#include "tcdefs.h"
#include "error.h"
#include "data.h"
#include "op.h"
#include "field.h"
#include "qdisc.h"
#include "police.h"
#include "tc.h"
#include "iflib.h"
#include "if.h"


static uint32_t handle;


static uint32_t new_handle(void)
{
    handle += 0x100000;
    return handle;
}


static void dump_handle(const char *tag,uint32_t handle)
{
    tc_more(" %s %x:%x:%x",tag,TC_U32_USERHTID(handle),TC_U32_HASH(handle),
      TC_U32_KEY(handle));
}


/*
 * Ugly hack: always have one hash table ready for use @@@
 */


static void next_hash(const FILTER *filter)
{
    __tc_filter_add(filter);
    dump_handle("handle",handle+0x100000);
    tc_more(" u32 divisor 1");
    tc_nl();
}


static int keys;
static uint32_t curr_handle;

static struct {
    int bits;
    int orig_offset;
    int offset;
    uint32_t value,mask;
} key[MAX_U32_KEYS];


static void begin_match(uint32_t handle)
{
    curr_handle = handle;
    keys = 0;
}


static void dump_match(const FILTER *filter)
{
    int i;

    __tc_filter_add(filter);
    if (TC_U32_HTID(curr_handle) == TC_U32_ROOT) tc_more(" u32");
    else {
	dump_handle("handle",curr_handle);
	tc_more(" u32");
	dump_handle("ht",TC_U32_HTID(curr_handle));
    }
    if (!keys) {
	tc_more(" match u32 0 0 at 0");
	return;
    }
    for (i = 0; i < keys; i++) {
	int shift;

	shift = 32-key[i].bits-(key[i].orig_offset-key[i].offset)*8;
	tc_more(" match u%d 0x%lx 0x%lx at %d",key[i].bits,
	  key[i].value >> shift,key[i].mask >> shift,key[i].orig_offset);
    }
}


/*
 * Return 0 if key contradicts previous keys.
 */


static int add_key(int bits,uint32_t value,uint32_t mask,int offset)
{
    int i;

    if (keys == MAX_U32_KEYS)
	errorf("too many u32 keys (max %d)",MAX_U32_KEYS);
    key[keys].orig_offset = offset;
    switch (bits) {
	case 8:
	    if (offset & 1) {
		offset--;
		value >>= 8;
		mask >>= 8;
	    }
	    /* fall through */
	case 16:
	    if (offset & 2) {
		offset -= 2;
		value >>= 16;
		mask >>= 16;
	    }
	    /* fall through */
	case 32:
	    break;
    }
    key[keys].bits = bits;
    key[keys].value = value;
    key[keys].mask = mask;
    key[keys].offset = offset;
    for (i = 0; i < keys; i++)
	if (key[i].offset == offset) {
	    if ((key[i].value^value) & key[i].mask & mask) return 0;
	    if (key[i].bits == bits &&
	      key[i].orig_offset == key[keys].orig_offset) {
		key[i].value |= key[keys].value;
		key[i].mask |= key[keys].mask;
		return 1;
	    }
	}
    keys++;
    return 1;
}


static int dump_or(const FILTER *filter,DATA d,uint32_t curr_handle);


static uint32_t __dump_link(const FILTER *filter)
{
    uint32_t next = new_handle();

    dump_handle("link",TC_U32_HTID(next));
    tc_nl();
    next_hash(filter);
    return next;
}


static void dump_link(const FILTER *filter,DATA d)
{
    uint32_t next;

    next = __dump_link(filter);
    dump_or(filter,d,next);
}


static void dump_offset(const FILTER *filter,DATA d)
{
    DATA off;

    dump_match(filter);
    off = d.op->b;
    if (!off.op) {
	if (off.type != dt_unum) dump_failed(off,"non-numeric offset");
	if (off.u.unum & 3)
	    errorf("offset must be a multiple of 4 (%ld)",
	      (unsigned long) off.u.unum);
	tc_more(" offset plus %d",off.u.unum);
    }
    else {
	uint32_t mask = 0xffff;
	int offset,shift,bytes;

	shift = 0;
	while (off.op->dsc == &op_shift_left ||
	  off.op->dsc == &op_shift_right) {
	    if (off.op->b.op || off.op->b.type != dt_unum)
		dump_failed(off,"bad shift");
	    shift = off.op->b.u.unum;
	    if (off.op->dsc == &op_shift_right) shift = -shift;
	    off = off.op->a;
	}
	if (off.op && off.op->dsc == &op_and) {
	    if (!off.op->a.op && off.op->a.type == dt_unum) {
		mask = off.op->a.u.unum;
		off = off.op->b;
	    }
	    else {
		if (off.op->b.op || off.op->b.type != dt_unum)
		    dump_failed(off,"non-constant and");
		mask = off.op->b.u.unum;
		off = off.op->a;
	    }
	    if (mask > 0xffff)
		errorf("offset mask must be < 0x10000 (not 0x%lx)",mask);
	}
	if (!off.op || off.op->dsc != &op_access)
	    dump_failed(off,"access expected");
	if (off.op->b.op || off.op->b.type != dt_unum)
	    dump_failed(off,"bad offset");
	offset = off.op->b.u.unum;
	bytes = off.op->c.u.unum/8;
	if (bytes == 1) {
	    shift -= 8;
	    mask = (mask & 0xff) << 8;
	}
	if (offset & 1) {
	    offset -= 1;
	    shift += 8;
	    if (mask & 0xff) errorf("offset adjustment removes non-zero bits");
	    mask >>= 8;
	}
	if (shift > 0) errorf("u32 can't left-shift (%d)",shift);
	if (shift < -15) errorf("u32 can't right-shift by %d",-shift);
	if (mask & ~(0xffff << -shift))
	    warnf("some non-zero bits of mask 0x%lx are always zero in value",
	      (unsigned long) mask,-shift);
	/*
	 * Brief history of mask and value:
	 *   iproute2/tc: mask read in natural order from line
 	 *   iproute2/tc: htons(mask) -> offmask
	 *   cls_u32: offset = ntohs(offmask & *(u16 *) ...) ...
	 *
	 * So we get:
	 *   offset = ntohs(htons(mask) & *(u16 *) ...)
	 *	    = mask & ntohs(*(u16 *) ...)
	 */
	tc_more(" offset at %d mask %04x shift %d eat",offset,mask,-shift);
    }
    dump_link(filter,d.op->a);
}


static int tc_match(int bits,uint32_t *value,uint32_t *mask,int *offset,
  int *bytes)
{
    int size = bits >> 3;
    int useful;

    if (*bytes < size || (*offset & (size-1))) return 1;
    useful = add_key(bits,*value,*mask,*offset);
    *value <<= bits;
    *mask <<= bits;
    *bytes -= size;
    *offset += size;
    return useful;
}


static int dump_eq(DATA d,DATA constant,DATA var)
{
    uint32_t value,mask,default_mask = 1;
    int offset,bytes,useful;

    if (constant.op || constant.type != dt_unum)
	dump_failed(d,"constant side isn't");
    if (!var.op) dump_failed(d,"variable side isn't");
    mask = ~0;
    while (var.op->dsc == &op_and) {
	if (var.op->a.type == dt_unum && !var.op->a.op) {
	    mask &= var.op->a.u.unum;
	    var = var.op->b;
	}
	else {
	    if (var.op->b.type != dt_unum || var.op->b.op)
		dump_failed(d,"non-constant and");
	    mask &= var.op->b.u.unum;
	    var = var.op->a;
	}
	default_mask = 0;
    }
    if (!var.op || var.op->dsc != &op_access) dump_failed(d,"access expected");
    if (var.op->b.op || var.op->b.type != dt_unum) dump_failed(d,"bad offset");
    offset = var.op->b.u.unum;
    bytes = var.op->c.u.unum/8;
    value = constant.u.unum;
    if (default_mask) mask = 0xffffffff >> (32-var.op->c.u.unum);
    if (bytes < 4 && (mask >> var.op->c.u.unum))
	warnf("mask 0x%lx larger than %d byte%s access",(unsigned long) mask,
	  bytes,bytes == 1 ? "" : "s");
    if (value & ~mask) return 0;
    value <<= 32-var.op->c.u.unum;
    mask <<= 32-var.op->c.u.unum;
    useful = tc_match(32,&value,&mask,&offset,&bytes);
    if (!tc_match(16,&value,&mask,&offset,&bytes)) useful = 0;
    if (!tc_match(8,&value,&mask,&offset,&bytes)) useful = 0;
    if (!tc_match(16,&value,&mask,&offset,&bytes)) useful = 0;
    if (!tc_match(8,&value,&mask,&offset,&bytes)) useful = 0;
    if (bytes) errorf("internal error: %d bytes left",bytes);
    return useful;
}


static int dump_other(const FILTER *filter,DATA d)
{
    if (d.op && d.op->dsc == &op_logical_and)
	error("internal error: && on left-hand side of &&");
    if (!d.op) {
	switch (d.type) {
	    case dt_unum:
		if (d.u.unum) return 1;
		return 0;
		/* earlier stages should have disappeard this ... */
	    case dt_decision:
		switch (d.u.decision.result) {
		    POLICE *drop_policer;

		    case dr_class:
			dump_match(filter);
			tc_add_classid(d.u.decision.class);
			break;
		    case dr_drop:
			dump_match(filter);
			drop_policer = get_drop_policer(filter->location);
			dump_police(drop_policer);
			tc_add_classid(get_any_class(filter->parent.qdisc));
			break;
		    case dr_continue:
			return 0; /* ignore "continue" */
		    default:
			dump_failed(d,"decision must be class or drop");
		}
		tc_nl();
		return 0;
	    default:
		break;
	}
    }
    else {
	 if (d.op->dsc == &op_logical_or) {
	    dump_match(filter);
	    dump_link(filter,d);
	    return 0;
	}
	if (d.op->dsc == &op_field_root)
	    error("u32 only supports field root 0");
	if (d.op->dsc == &op_offset) {
	    dump_offset(filter,d);
	    return 1;
	}
	if (d.op->dsc == &op_eq) {
	    if (d.op->a.type == dt_unum && !d.op->a.op)
		return dump_eq(d,d.op->a,d.op->b);
	    if (d.op->b.type == dt_unum && !d.op->b.op)
		return dump_eq(d,d.op->b,d.op->a);
	}
    }
    dump_failed(d,"not recognized by dump_other");
    return 0; /* not reached */
}


static int dump_and(const FILTER *filter,DATA d,uint32_t *curr_handle)
{
    /* @@@ check: 0-based key allowed ? also: check key overflow */
    if (d.op && d.op->dsc == &op_logical_or)
	error("internal error: || on left-hand side of ||");
    (*curr_handle)++;
    begin_match(*curr_handle);
    while (d.op && d.op->dsc == &op_logical_and) {
	if (d.op->a.op &&
	  (d.op->a.op->dsc == &op_count || d.op->a.op->dsc == &op_conform)) {
abort();
	}
	if (!dump_other(filter,d.op->a)) return 0;
	d = d.op->b;
    }
    return dump_other(filter,d);
}


static int dump_or(const FILTER *filter,DATA d,uint32_t curr_handle)
{
    int useful = 1;

    while (d.op && d.op->dsc == &op_logical_or) {
	if (dump_and(filter,d.op->a,&curr_handle)) useful = 1;
	d = d.op->b;
    }
    return dump_and(filter,d,&curr_handle) || useful;
}


static int num_continues(DATA d)
{
    if (d.type == dt_decision) return d.u.decision.result == dr_continue;
    if (!d.op) return 0;
    return num_continues(d.op->a)+num_continues(d.op->b)+num_continues(d.op->c);
}


static int has_continue(DATA d)
{
    while (d.op && (d.op->dsc == &op_offset || d.op->dsc == &op_logical_and)) {
	if (d.op->dsc == &op_offset) d = d.op->a;
	else {
	    if (d.op->a.type == dt_decision)
		return d.op->a.u.decision.result == dr_continue;         
	    d = d.op->b;
	}
    }
    return d.type == dt_decision && d.u.decision.result == dr_continue;
}


static int continue_at_end(DATA d)
{
    int n,in_continue,this;

    n = num_continues(d);
    in_continue = 0;
    while (d.op && (d.op->dsc == &op_logical_or || d.op->dsc == &op_offset)) {
	if (d.op->dsc == &op_offset) d = d.op->a;
	else {
	    this = has_continue(d.op->a);
	    if (!this && in_continue) return 0;
	    in_continue = this;
	    if (this) n--;
	    d = d.op->b;
	}
    }
    this = has_continue(d);
    if (!this && in_continue) return 0;
    if (this) n--;
    return !n;
}


void dump_if_u32(const FILTER *filter)
{
    QDISC *qdisc = filter->parent.qdisc;
    DATA d;
    int i;

    tc_pragma(filter->params);
    for (i = 0; i < 2; i++) {
	if (!i) {
	    d = data_clone(qdisc->if_expr);
	    d = op_binary(&op_logical_or,d,data_decision(dr_continue,NULL));
	}
	else {
	    if (!(qdisc->dsc->flags & QDISC_HAS_DEFAULT))
		lerrorf(qdisc->location,
		  "qdisc \"%s\" does not provide default class",
		  qdisc->dsc->name);
	    d = qdisc->if_expr;
	    qdisc->dsc->default_class(&d,qdisc);
	}
	iflib_arith(&d);
	iflib_not(&d);
	iflib_reduce(&d);
	iflib_normalize(&d);
	iflib_actions(filter->location,&d);
	iflib_offset(&d,1);
	if (d.op && d.op->dsc == &op_field_root &&
	  d.op->b.u.field_root->offset_group == 0)
	    d = d.op->a;
	if (!d.op && d.type == dt_unum && !d.u.unum) return;
	if (!i && continue_at_end(d)) break;
    }
    handle = 0;
    next_hash(filter);
    (void) dump_or(filter,d,TC_U32_ROOT);
}
